
import json, os, glob, subprocess
from collections import defaultdict

from planutils import settings

PACKAGES = {}

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def check_package(target, manifest):
    assert os.path.exists(manifest), "Error: Manifest must be defined for %s" % target
    with open(manifest, 'r') as f:
        config = json.load(f)
    for key in ['name', 'description', 'dependencies', 'install-size']:
        assert key in config, "Error: Manifest for %s must include '%s'" % (base, key)


for conf_file in glob.glob(os.path.join(CUR_DIR, 'packages', '*')):
    base = os.path.basename(conf_file)
    if base not in ['README.md', 'TEMPLATE']:
        assert base not in PACKAGES, "Error: Duplicate package config -- %s" % base
        check_package(base, os.path.join(conf_file, 'manifest.json'))
        with open(os.path.join(conf_file, 'manifest.json'), 'r') as f:
            config = json.load(f)
        PACKAGES[base] = config
        PACKAGES[base]['runnable'] = os.path.exists(os.path.join(conf_file, 'run'))


def check_installed(target):
    return target in settings.load()['installed']


def uninstall(targets):

    for target in targets:
        if target not in PACKAGES:
            print("Error: Package not found -- %s" % target)
            return

    to_check = []
    for target in targets:
        if check_installed(target):
            to_check.append(target)
        else:
            print("%s isn't installed." % target)

    if not to_check:
        return

    s = settings.load()
    # map a package to all those that depend on it
    dependency_mapping = defaultdict(set)
    for p in s['installed']:
        for dep in PACKAGES[p]['dependencies']:
            dependency_mapping[dep].add(p)

    # compute all the packages that will be removed
    to_remove = set()
    while to_check:
        check = to_check.pop(0)
        to_remove.add(check)
        to_check.extend(list(dependency_mapping[check]))

    print("\nAbout to remove the following packages: %s" % ', '.join(to_remove))
    if input("  Proceed? [y/N] ").lower() in ['y', 'yes']:
        for package in to_remove:
            print ("Uninstalling %s..." % package)
            subprocess.call('./uninstall', cwd=os.path.join(CUR_DIR, 'packages', package))
            print ("Finished uninstalling %s" % package)
            s['installed'].remove(package)

    # Search for any packages that may be no longer required
    dependency_mapping = defaultdict(set)
    for p in PACKAGES:
        for dep in PACKAGES[p]['dependencies']:
            dependency_mapping[dep].add(p)

    possible_deletions = [d for p in to_remove for d in PACKAGES[p]['dependencies']]
    while possible_deletions:
        package = possible_deletions.pop(0)
        # Consider removing if (1) it's installed; and (2) nothing that requires this is installed
        if (package in s['installed']) and (not any([p in s['installed'] for p in dependency_mapping[package]])):
            print("\nPackage may no longer be required: %s" % package)
            if input("  Remove? [y/N] ").lower() in ['y', 'yes']:
                print ("Uninstalling %s..." % package)
                subprocess.call('./uninstall', cwd=os.path.join(CUR_DIR, 'packages', package))
                print ("Finished uninstalling %s" % package)
                s['installed'].remove(package)
                possible_deletions.extend(PACKAGES[package]['dependencies'])

    settings.save(s)


def package_list():
    print("\nInstalled:")
    installed = set(settings.load()['installed'])
    for p in installed:
        print("  %s: %s" % (p, PACKAGES[p]['name']))

    print("\nAvailable:")
    for p in PACKAGES:
        if p not in installed:
            print("  %s: %s" % (p, PACKAGES[p]['name']))
    print()

def upgrade():
    s = settings.load()
    for package in s['installed']:
        print("Upgrading %s..." % package)
        subprocess.call('./uninstall', cwd=os.path.join(CUR_DIR, 'packages', package))
        subprocess.call('./install', cwd=os.path.join(CUR_DIR, 'packages', package))

def install(targets, forced=False):
    for target in targets:
        if target not in PACKAGES:
            print("Error: Package not found -- %s" % target)
            return False

    # Compute all those that will need to be installed
    to_check = []
    for target in targets:
        if check_installed(target):
            print("%s is already installed." % target)
        else:
            to_check.append(target)

    done = set()
    to_install = []
    while to_check:
        check = to_check.pop(0)
        if check not in done:
            done.add(check)
            if not check_installed(check):
                to_install.append(check)
                to_check.extend(PACKAGES[check]['dependencies'])

    to_install.reverse()

    if to_install:
        to_install_desc = ["%s (%s)" % (pkg, PACKAGES[pkg]['install-size']) for pkg in to_install]
        print("\nAbout to install the following packages: %s" % ', '.join(to_install_desc))

        if forced:
            user_response = True
        else:
            user_response = input("  Proceed? [Y/n] ").lower() in ['', 'y', 'yes']

        if user_response:
            installed = []
            for package in to_install:
                package_path = os.path.join(CUR_DIR, 'packages', package)
                print("Installing %s..." % package)
                try:
                    installed.append(package)
                    subprocess.check_call('./install', cwd=package_path, shell=True)
                    size = subprocess.check_output('du -sh .',
                                                   cwd=package_path,
                                                   shell=True,
                                                   encoding='utf-8').split('\t')[0]
                    print("Finished installing %s (size: %s)" % (package, size))
                except subprocess.CalledProcessError:
                    print("\nError installing %s. Rolling back changes..." % package)
                    for p in installed:
                        subprocess.call('./uninstall', cwd=os.path.join(CUR_DIR, 'packages', p))
                    return False

            s = settings.load()
            s['installed'].extend(installed)
            settings.save(s)
            return True
        else:
            print("Aborting installation.")
    else:
        print("Nothing left to install.")

    return False





import json, os

# This should eventually be changed once the prefix is customizable
PLANUTILS_PREFIX = os.path.join(os.path.expanduser('~'), '.planutils')

SETTINGS_FILE = os.path.join(PLANUTILS_PREFIX, 'settings.json')

def load():
    with open(SETTINGS_FILE, 'r') as f:
        settings = json.loads(f.read())
    return settings

def save(s):
    with open(SETTINGS_FILE, 'w') as f:
        f.write(json.dumps(s))




import setuptools

# Will force a check on the packages setup
from planutils.package_installation import PACKAGES

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='planutils',
      version='0.2.10',
      description='General library for setting up linux-based environments for developing, running, and evaluating planners.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/AI-Planning/planutils',
      author='',
      author_email='',
      license='MIT',
      packages=['planutils'],
      scripts=['bin/planutils'],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: POSIX :: Linux",
      ],
      python_requires='>=3.6',
      include_package_data=True,
      zip_safe=False)




import argparse, os

from planutils import settings
from planutils.package_installation import PACKAGES


def setup():

    if setup_done():
        print("\nError: planutils is already setup. Setting up again will wipe all cached packages and settings.")
        if input("  Proceed? [y/N] ").lower() in ['y', 'yes']:
            os.system("rm -rf %s" % os.path.join(os.path.expanduser('~'), '.planutils'))
        else:
            return

    CUR_DIR = os.path.dirname(os.path.abspath(__file__))

    print("\nCreating ~/.planutils...")
    os.mkdir(os.path.join(os.path.expanduser('~'), '.planutils'))
    os.mkdir(os.path.join(os.path.expanduser('~'), '.planutils', 'bin'))

    settings.save({
        'installed': []
    })

    os.symlink(os.path.join(CUR_DIR, 'packages'),
               os.path.join(os.path.expanduser('~'), '.planutils', 'packages'))

    print("Adding bin folder to path (assuming ~/.bashrc exists)...")
    with open(os.path.join(os.path.expanduser('~'), '.bashrc'), "a+") as f:
        f.write("export PLANUTILS_PREFIX=\"~/.planutils\"\n")
        f.write("export PATH=\"$PLANUTILS_PREFIX/bin:$PATH\"\n")

    print("Installing package scripts...")
    for p in PACKAGES:
        if PACKAGES[p]['runnable']:
            script  = "#!/bin/bash\n"
            script += "if $(planutils check-installed %s)\n" % p
            script += "then\n"
            script += "  ~/.planutils/packages/%s/run $@\n" % p
            script += "else\n"
            script += "  echo\n"
            script += "  echo 'Package not installed!'\n"
            script += "  read -r -p \"  Download & install? [Y/n] \" varchoice\n"
            script += "  varchoice=${varchoice,,}\n" # tolower
            script += "  if ! [[ \"$varchoice\" =~ ^(no|n)$ ]]\n"
            script += "  then\n"
            script += "    if planutils install %s;\n" % p
            script += "    then\n"
            script += "      echo 'Successfully installed %s!'\n" % p
            script += "      echo\n"
            script += "      echo \"Original command: %s $@\"\n" % p
            script += "      read -r -p \"  Re-run command? [Y/n] \" varchoice\n"
            script += "      varchoice=${varchoice,,}\n" # tolower
            script += "      if ! [[ \"$varchoice\" =~ ^(no|n)$ ]]\n"
            script += "      then\n"
            script += "        ~/.planutils/packages/%s/run $@\n" % p
            script += "      fi\n"
            script += "    fi\n"
            script += "  fi\n"
            script += "  echo\n"
            script += "fi\n"
            with open(os.path.join(os.path.expanduser('~'), '.planutils', 'bin', p), 'w') as f:
                f.write(script)
            os.chmod(os.path.join(os.path.expanduser('~'), '.planutils', 'bin', p), 0o0755)


    print("\nAll set! Be sure to start a new bash session or update your PATH variable to include ~/.planutils/bin\n")

def setup_done():
    return os.path.exists(os.path.join(os.path.expanduser('~'), '.planutils'))


def main():
    parser = argparse.ArgumentParser(prog="planutils")
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')

    parser_install = subparsers.add_parser('install', help='install package(s) such as a planner')
    parser_install.add_argument('-f', '--force', help='force installation', action='store_true')
    parser_install.add_argument('package', help='package name', nargs='+')

    parser_uninstall = subparsers.add_parser('uninstall', help='uninstall package(s)')
    parser_uninstall.add_argument('package', help='package name', nargs='+')

    parser_checkinstalled = subparsers.add_parser('check-installed', help='check if a package is installed')
    parser_checkinstalled.add_argument('package', help='package name')

    parser_list = subparsers.add_parser('list', help='list the available packages')
    parser_setup = subparsers.add_parser('setup', help='setup planutils for current user')
    parser_upgrade = subparsers.add_parser('upgrade', help='upgrade all of the installed packages')

    args = parser.parse_args()

    if 'setup' == args.command:
        setup()
    elif not setup_done():
        print("\nPlease run 'planutils setup' before using utility.\n")

    elif 'check-installed' == args.command:
        from planutils.package_installation import check_installed
        exit({True:0, False:1}[check_installed(args.package)])

    elif 'install' == args.command:
        from planutils.package_installation import install
        exit({True:0, False:1}[install(args.package, args.force)])

    elif 'uninstall' == args.command:
        from planutils.package_installation import uninstall
        uninstall(args.package)

    elif 'list' == args.command:
        from planutils.package_installation import package_list
        package_list()

    elif 'upgrade' == args.command:
        from planutils.package_installation import upgrade
        upgrade()

    else:
        parser.print_help()
